"""Top level package: make the user-facing functions top-level, rest accessed as submodules."""

__version__ = "0.6.0"

from .user_funcs import (
    log_forward_pass,
    show_model_graph,
    get_model_metadata,
    validate_saved_activations,
    validate_batch_of_models_and_inputs,
)
from .model_history import ModelHistory
from .tensor_log import TensorLogEntry, RolledTensorLogEntry
